import argparse
import os
import sys
import string
import random
import requests
import logging
import colorlog

from concurrent.futures import ThreadPoolExecutor

from aicoder.AICoderFull import AICoderFull
from aicoder.AICoderPartial import AICoderPartial

# Logger setup
handler = colorlog.StreamHandler()
handler.setFormatter(colorlog.ColoredFormatter('%(log_color)s%(levelname)s:%(message)s'))
logger = colorlog.getLogger('file_generator')
if not logger.hasHandlers():
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

# Function to generate a random string
def get_random_string(length):
    letters = string.ascii_letters
    return ''.join(random.choice(letters) for _ in range(length))

# Function to generate a full program
def full_generator(prompt: str, api_key: str, github_token: str, model: str="gpt-4"):
    if os.path.isdir("generated"):
        os.system("rm -rf generated")
    aicoderfull = AICoderFull(prompt, api_key, github_token, model)
    aicoderfull.generate_program()

# Function to generate a file from templates
def file_generator(prompt: str, api_key: str, github_token: str, model: str, template_files: list, orig_branch: str, new_branch: str):
    aicoderpartial = AICoderPartial(prompt, api_key, github_token, model)
    aicoderpartial.gen_file_from_templates(template_files, prompt, orig_branch, new_branch)

# Function to enhance a file
def enhancer(prompt: str, api_key: str, github_token: str, model: str, file_path: str, orig_branch: str, new_branch: str):
    aicoderpartial = AICoderPartial(prompt, api_key, github_token, model)
    aicoderpartial.modify_file(file_path, prompt, orig_branch, new_branch)

# Function to add comments to a file
def add_comments(prompt: str, api_key: str, github_token: str, model: str, file_path: str, orig_branch: str, new_branch: str):
    final_prompt = "Given some code, add useful comments to the file inside the code to easily understand what it does."
    if prompt:
        final_prompt = "\n" + prompt
    aicoderpartial = AICoderPartial(final_prompt, api_key, github_token, model)
    aicoderpartial.modify_file(file_path, final_prompt, orig_branch, new_branch)

# Function to check for security vulnerabilities in a file
def check_security(prompt: str, api_key: str, github_token: str, model: str, file_path: str, orig_branch: str, new_branch: str):
    final_prompt = "Given some code, search for security vulnerabilities and potential backdoors on it and if any, create a new code fixing it (if possible). Return only the fixed code with the comments. If you changed anything, add at the end of the code comments explaining the changes."
    if prompt:
        final_prompt = "\n" + prompt
    aicoderpartial = AICoderPartial(final_prompt, api_key, github_token, model)
    aicoderpartial.modify_file(file_path, final_prompt, orig_branch, new_branch)

# Function to optimize a file
def optimize_file(prompt: str, api_key: str, github_token: str, model: str, file_path: str, orig_branch: str, new_branch: str):
    final_prompt = "Given some code, optimize it without loosing functionality. Return a code that keeping the same functionality and respecting the class and function names, it's more clean, simple, don't have duplice code, don't have code smells and is more efficient if possible. If you cannot improve the code just return the same code. If you changed anything, add at the end of the code comments explaining the changes. Be careful with chunks of code that looks similar but aren't the same, even  ifthe code can be improved the different functionalities must be keeped."
    if prompt:
        final_prompt = "\n" + prompt
    aicoderpartial = AICoderPartial(final_prompt, api_key, github_token, model)
    aicoderpartial.modify_file(file_path, final_prompt, orig_branch, new_branch)

# Function to add common arguments to a parser
def add_common_arguments(parser):
    parser.add_argument('--api-key', default=None, type=str, help='Input API key')
    parser.add_argument('--model', default="gpt-4", type=str, help='Model to use')

# Main function
def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="mode")  # Create subparsers

    # Create subparsers
    full_generator_parser = subparsers.add_parser('full-generator', help='Generate a full program from the prompt.')
    file_generator_parser = subparsers.add_parser('file-generator', help='Generate a file based on others as templates.')
    file_enhancer_parser = subparsers.add_parser('file-enhancer', help='Enhance a file based on the given prompt.')
    file_commentor_parser = subparsers.add_parser('file-comments', help='Ask to add comments to a file.')
    file_security_parser = subparsers.add_parser('file-security', help='Ask to search security vulnerabilities and backdoors in a file.')
    file_optimizer_parser = subparsers.add_parser('file-optimizer', help='Ask to optimize a file in terms of clean code and efficiency.')

    # Add common arguments to each subparser
    for subparser in [full_generator_parser, file_generator_parser, file_enhancer_parser, file_commentor_parser, file_security_parser, file_optimizer_parser]:
        add_common_arguments(subparser)

    # Add specific arguments to each subparser
    full_generator_parser.add_argument('--prompt', default=None, type=str, help='Input prompt. It can be string, a file path or a url.')
    
    file_generator_parser.add_argument('--template-files', nargs='+', help='List of template files')
    file_generator_parser.add_argument('--github-token', default=None, type=str, help='Github token')
    file_generator_parser.add_argument('--orig-branch', default=None, type=str, help='Branch to upload the changes.')
    file_generator_parser.add_argument('--to-branch', default=None, type=str, help='Branch to send the PR to.')
    file_generator_parser.add_argument('--prompt', default=None, type=str, help='Input prompt. It can be string, a file path or a url.')
    
    file_enhancer_parser.add_argument('--path', type=str, help='Path to file to enhance')
    file_enhancer_parser.add_argument('--github-token', default=None, type=str, help='Github token')
    file_enhancer_parser.add_argument('--orig-branch', default=None, type=str, help='Branch to upload the changes.')
    file_enhancer_parser.add_argument('--to-branch', default=None, type=str, help='Branch to send the PR to.')
    file_enhancer_parser.add_argument('--prompt', default=None, type=str, help='Input prompt. It can be string, a file path or a url.')
    
    file_commentor_parser.add_argument('--path', type=str, help='Path to file to enhance')
    file_commentor_parser.add_argument('--github-token', default=None, type=str, help='Github token')
    file_commentor_parser.add_argument('--orig-branch', default=None, type=str, help='Branch to upload the changes.')
    file_commentor_parser.add_argument('--to-branch', default=None, type=str, help='Branch to send the PR to.')
    
    file_security_parser.add_argument('--path', type=str, help='Path to file to enhance')
    file_security_parser.add_argument('--github-token', default=None, type=str, help='Github token')
    file_security_parser.add_argument('--orig-branch', default=None, type=str, help='Branch to upload the changes.')
    file_security_parser.add_argument('--to-branch', default=None, type=str, help='Branch to send the PR to.')
    
    file_optimizer_parser.add_argument('--path', type=str, help='Path to file to enhance')
    file_optimizer_parser.add_argument('--github-token', default=None, type=str, help='Github token')
    file_optimizer_parser.add_argument('--orig-branch', default=None, type=str, help='Branch to upload the changes.')
    file_optimizer_parser.add_argument('--to-branch', default=None, type=str, help='Branch to send the PR to.')
    file_optimizer_parser.add_argument('--prompt', default=None, type=str, help='Input prompt. It can be string, a file path or a url. In this case, it\'s optional')
    
    # Parse arguments
    args = parser.parse_args()
    prompt = os.environ.get("INPUT_PROMPT") or (args.prompt if hasattr(args,"prompt") else None)
    api_key = os.environ.get("INPUT_API_KEY") or (args.api_key if hasattr(args,"api_key") else None)
    model = os.environ.get("INPUT_MODEL") or (args.model if hasattr(args,"model") else None)

    if args.mode in ["file-generator", "file-enhancer", "file-comments", "file-security", "file-optimizer"]:
        github_token = os.environ.get("GITHUB_TOKEN") or args.github_token
        orig_branch = os.environ.get("ORIGIN_BRANCH") or args.orig_branch
        to_branch = os.environ.get("TO_BRANCH") or args.to_branch

        if args.mode == "file-generator":
            template_files = list(os.environ.get("TEMPALTE_FILES").split(",")) or args.template_files
        
        elif args.mode == "file-enhancer":
            modify_file_path = os.environ.get("MODIFY_FILE_PATH") or args.path
        
        elif args.mode in ["file-comments", "file-security", "file-optimizer"]:
            check_path = os.environ.get("CHECK_PATH") or args.path
    
    if not prompt and not args.mode in ["file-comments", "file-security", "file-optimizer"]:
        logger.error("Error: Missing prompt.")
        sys.exit(1)
    
    if not api_key:
        logger.error("Error: Missing API key.")
        sys.exit(1)
    
    # If prompt is file path, read it
    if prompt and os.path.isfile(prompt):
        logger.debug(f"Prompt read from {prompt}")
        with open(prompt, "r") as f:
            prompt = f.read()
    
    # If prompt is url, download it
    elif prompt and prompt.startswith("http"):
        logger.debug(f"Prompt downloaded from {prompt}")
        prompt = requests.get(prompt).text
    
    # If full generator mode
    if args.mode == 'full-generator':
        full_generator(prompt, api_key, github_token, model)
        exit(0)
    
    ### Checks for partial generators
    
    if not github_token:
        logger.warning("No Github token. PRs won't be created automatically.")
    
    if github_token and not orig_branch:
        logger.error("Error: No orig branch. PRs won't be created automatically. Stopping.")
        exit(1)
    
    if github_token and not to_branch:
        logger.error("Error: No to branch. PRs won't be created automatically. Stopping.")
        exit(1)

    # If file enhancer mode
    if args.mode == 'file-enhancer':
        enhancer(prompt, api_key, github_token, model, modify_file_path, orig_branch, to_branch)
        exit(0)
    
    # If file enhancer mode
    elif args.mode == 'file-generator':
        file_generator(prompt, api_key, github_token, model, template_files, orig_branch, to_branch)
        exit(0)
    
    # If file commentor mode
    elif args.mode in ['file-comments', 'file-security', 'file-optimizer']:
        # pompt is optional for these modes and usually None
        if args.mode == 'file-comments':
            f = add_comments
        elif args.mode == 'file-security':
            f = check_security
        elif args.mode == 'file-optimizer':
            f = optimize_file
        
        # Check if path is file or folder
        if os.path.isfile(check_path):
            logger.debug(f"Given path {check_path} is a file.")
            f(prompt, api_key, github_token, model, check_path, orig_branch, to_branch)
        
        elif os.path.isdir(check_path):
            logger.debug(f"Given path {check_path} is a folder.")
            with ThreadPoolExecutor(max_workers=3) as executor:
                for root, dirs, files in os.walk(check_path):
                    for file in files:
                        logger.debug(f"Checking file: {check_path}")
                        executor.submit(f, prompt, api_key, github_token, model, os.path.join(root, file), orig_branch, to_branch)
        exit(0)

    
if __name__ == "__main__":
    main()
